import abc
from typing import TYPE_CHECKING, TypeVar

from . import logger
from .geometry import CoordinatesType, Region
from .utils import ABC, argument_guard, image_utils

if TYPE_CHECKING:
    from PIL.Image import Image

    from .geometry import Point

__all__ = ("EyesScreenshot",)

Self = TypeVar("Self", bound="EyesScreenshot")  # typedef


class EyesScreenshot(ABC):
    """
    Base class for handling screenshots.
    """

    def __init__(self, image):
        # type: (Image) -> None
        argument_guard.not_none(image)
        self._image = image

    def __hash__(self):
        return hash(image_utils.get_base64(self._image))

    @property
    def image(self):
        # type: () -> Image
        return self._image

    @abc.abstractmethod
    def sub_screenshot(self, region, throw_if_clipped=False):
        # type: (Self, Region, bool) -> Self
        """
        Returns a part of the screenshot based on the given region.

        :param region:           The region for which we should get the sub screenshot.
        :param throw_if_clipped: Throw an EyesException if the region is not
                                 fully contained in the screenshot.
        :return: A screenshot instance containing the given region.
        """

    @abc.abstractmethod
    def convert_location(self, location, from_, to):
        # type: (Point, CoordinatesType, CoordinatesType) -> Point
        """
        Converts a location's coordinates with the `from_` coordinates type
        to the `to` coordinates type.

        :param location:  The location which coordinates needs to be converted.
        :param from_:     The current coordinates type for `location`.
        :param to:        The target coordinates type for `location`.
        :return A new location which is the transformation of `location` to
                the `to` coordinates type.
        """

    @abc.abstractmethod
    def location_in_screenshot(self, location, coordinates_type):
        # type: (Point, CoordinatesType) -> Point
        """
        Calculates the location in the screenshot of the location given as
        parameter.

        :param location:         The location as coordinates inside the current frame.
        :param coordinates_type: The coordinates type of `location`.
        :return: The corresponding location inside the screenshot,
                 in screenshot as-is coordinates type.
        :raise: `OutOfBoundsException` If the location is
                not inside the frame's region in the screenshot.
        """

    @abc.abstractmethod
    def intersected_region(self, region, coordinates_type):
        # type: (Region, CoordinatesType) -> Region
        """
        Get the intersection of the given region with the screenshot.

        :param region:          The region to intersect.
        :param coordinates_type: The coordinates type of `region`.
        :return The intersected region, in `coordinates_type` coordinates.
        """

    def convert_region_location(self, region, from_, to):
        # type: (Region, CoordinatesType, CoordinatesType) -> Region
        """
        Converts a region's location coordinates with the `from_`
        coordinates type to the `to` coordinates type.

        :param region: The region which location's coordinates needs to be converted.
        :param from_:  The current coordinates type for `region`.
        :param to:     The target coordinates type for `region`.
        :return:       A new region which is the transformation of `region` to
                       the `to` coordinates type.
        """
        argument_guard.not_none(region)
        argument_guard.is_a(region, Region)
        logger.debug("convert_region_location({}, {}, {})".format(region, from_, to))

        if region.is_size_empty:
            return Region.EMPTY()

        argument_guard.not_none(from_)
        argument_guard.not_none(to)

        updated_location = self.convert_location(region.location, from_, to)
        logger.debug("updated_location: {}".format(updated_location))
        return Region(
            updated_location.x,
            updated_location.y,
            region.width,
            region.height,
            coordinates_type=to,
        )

    @property
    def image_region(self):
        # type: () -> Region
        return Region(
            0,
            0,
            self._image.size.width,
            self._image.size.height,
            CoordinatesType.SCREENSHOT_AS_IS,
        )



import typing
from enum import Enum

import attr

from applitools.common.utils.json_utils import JsonInclude

from . import utils

if typing.TYPE_CHECKING:
    from typing import List, Text


class TraceLevel(Enum):
    Debug = 0
    Info = 1
    Notice = 2
    Warn = 3
    Error = 4


@attr.s
class ClientEvent(object):
    level = attr.ib(type=TraceLevel, metadata={JsonInclude.THIS: True})
    event = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    timestamp = attr.ib(
        factory=utils.current_time_in_iso8601,
        metadata={JsonInclude.THIS: True},
    )  # type: Text


@attr.s
class LogSessionsClientEvents(object):
    events = attr.ib(
        factory=list, metadata={JsonInclude.THIS: True}
    )  # type List[ClientEvent]



from __future__ import absolute_import

import hashlib
import inspect
import os
import typing

import attr

from .compat import parse_qs, urlencode, urlsplit, urlunsplit

"""
General purpose utilities.
"""


if typing.TYPE_CHECKING:
    from typing import Any, Callable, List, Optional, Text

    T = typing.TypeVar("T")


def use_default_if_none_factory(default_obj, obj):
    def default(attr_name):
        val = getattr(obj, attr_name)
        if val is None:
            return getattr(default_obj, attr_name)
        return val

    return default


def cached_property(f):
    # type: (Callable) -> Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def get_sha256_hash(content):
    return hashlib.sha256(content).hexdigest()


def set_query_parameter(url, param_name, param_value):
    """Given a URL, set or replace a query parameter and return the
    modified URL.

    >>> set_query_parameter('http://example.com?foo=bar&biz=baz', 'foo', 'stuff')
    'http://example.com?foo=stuff&biz=baz'

    """
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    query_params[param_name] = [param_value]
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def proxy_to(proxy_obj_name, fields=None):
    # type: (Text, Optional[List[Text]]) -> Callable
    """
    Adds to decorated class __getter__ and __setter__ methods that allow to access
    attributes from proxy_object in the parent class.

    :param proxy_obj_name: The name of the proxy object that has decorated class.
    :param fields:
        Fields which should be accessible in parent object from the proxy object.
    """

    def __getattr__(self, name):
        _fields = fields or self._proxy_to_fields or []
        if name in _fields:
            proxy_obj = getattr(self, proxy_obj_name)
            return getattr(proxy_obj, name)
        module_with_class = "{}::{}".format(
            inspect.getfile(self.__class__), self.__class__.__name__
        )
        raise AttributeError("{} has not attr `{}`".format(module_with_class, name))

    def _setattr(cls):
        def __setattr__(self, key, value):
            _fields = fields or self._proxy_to_fields or []
            if key in _fields:
                proxy_obj = getattr(self, proxy_obj_name)
                setattr(proxy_obj, key, value)
            elif self.__class__ is cls:
                # if current class is the same that we've decorated
                super(self.__class__, self).__setattr__(key, value)
            else:
                # to prevent recursion error if not the current class is decorated one
                self.__dict__[key] = value

        return __setattr__

    def __dir__(self):
        _fields = fields or self._proxy_to_fields or []
        origin_fields = dir(self.__class__)
        return list(origin_fields) + _fields

    def dec(cls):
        cls._proxy_to_fields = None
        cls.__getattr__ = __getattr__
        cls.__setattr__ = _setattr(cls)
        cls.__dir__ = __dir__
        return cls

    return dec


def all_fields(obj):
    # type: (Any) -> List[Text]
    """Get all public fields from the object

    Args:
        obj: any kind object

    Returns:
        list of attributes and methods names
    """
    if attr.has(obj):
        return list(attr.fields_dict(obj).keys())
    else:
        return [f for f in vars(obj) if not f.startswith("_")]


def all_attrs(obj):
    # type: (Any) -> List[Text]
    """Get all public attributes from the object. Methods and fields.

    Args:
        obj: any kind object

    Returns:
        list of attributes and methods names
    """
    return [f for f in dir(obj) if not f.startswith("_")]


def get_env_with_prefix(env_name, default=None):
    # type: (Text, Optional[Text]) -> Optional[Text]
    """
    Takes name of ENV variable, check if exists origin and with list of prefixes
    """
    prefixes_to_check = ["bamboo"]
    try:
        return os.environ[env_name]
    except KeyError:
        for prefix in prefixes_to_check:
            name = "{}_{}".format(prefix, env_name)
            value = os.getenv(name)
            if value:
                return value
    return default


def counted(f):
    """
    Decorator that tracks how many times the function is called
    """

    def wrapped(*args, **kwargs):
        wrapped.calls += 1
        return f(*args, **kwargs)

    wrapped.calls = 0
    return wrapped


class DynamicEnumGetter(object):
    """
    Allow to use methods as values in Enum classes
    """

    def __init__(self, fget=None):
        self.fget = fget
        self.__doc__ = fget.__doc__

    def __get__(self, instance, ownerclass=None):
        return self.fget(ownerclass)



class DictAccessMixin(object):
    """Make dict-like object from attrs class"""

    def __getitem__(self, item):
        if isinstance(item, int):
            item = self.__slots__[item]
        if item not in self.__slots__:
            raise KeyError
        return getattr(self, item)



from applitools.common.validators import is_list_or_tuple

from .compat import iteritems


def get_name_from_param(param):
    for var_name, var in iteritems(locals()):
        if var is param:
            return var_name


def not_none(param, exc_info=None):
    """
    Fails if the input parameter is None
    """
    if param is None:
        if exc_info:
            raise exc_info
        raise ValueError("{} is None".format(get_name_from_param(param)))


def not_list_or_tuple(param, exc_info=None):
    """
    Fails if param is not a list or tuple
    """
    if not is_list_or_tuple(param):
        if exc_info:
            raise exc_info
        raise ValueError("{} is not list or tuple".format(get_name_from_param(param)))


def not_none_or_empty(param, exc_info=None):
    """
    Fails if the input parameter string is null or empty.
    """
    not_none(param)
    if not param:
        if exc_info:
            raise exc_info
        raise ValueError("{} is empty".format(get_name_from_param(param)))


def greater_than_or_equal_to_zero(param, *args):
    # type: (int, *int, Exception) -> None
    """
    Fails if the input integer parameter is negative.
    """
    if args:
        # adaptation for attr library
        param = args[1]
    if 0 >= param:
        raise ValueError("{} < 0".format(get_name_from_param(param)))


# TODO: update after resolving issue
def greater_than_zero(param, exc_info=None):
    # type: (int) -> None
    if 0 >= param:
        if exc_info:
            raise exc_info
        raise ValueError("{} < 1".format(get_name_from_param(param)))


def is_valid_state(is_valid, message):
    """
    Fails if is_valid is false.
    """
    if not is_valid:
        raise ValueError(message)


def is_a(param, klass, exc_info=None):
    if not isinstance(param, klass):
        if exc_info:
            raise exc_info
        raise ValueError("{} is not instance of {}".format(param, klass))


def is_in(param, klass, exc_info=None):
    if not any(isinstance(param, kls) for kls in klass):
        if exc_info:
            raise exc_info
        raise ValueError("{} is not instance of {}".format(param, klass))


def are_(params, klass, exc_info=None):
    for param in params:
        is_a(param, klass, exc_info)
